//=============================================================================
//||                             DirixleJacobi                               ||
//||                                                                         ||
//||                               MainGPU.cu                                ||
//||                                                                         ||
//||                               by Dampir                                 ||
//=============================================================================

#include "CudaHeaders.cuh"
//#include <thrust/extrema.h>

#define BLOCK_SIZE 16
#define CntRunKernel 10

__device__  unsigned int exact = 1000;
__device__  float ex;

cudaCompCapability chk_compute_copability()
{
	cudaCompCapability CC;
	int devID;
    	cudaDeviceProp props;   
    	cudaGetDevice(&devID);  // get number of SMs on this GPU
    	cudaGetDeviceProperties(&props, devID);
	CC.major = props.major;
	CC.minor = props.minor;
	return CC;
}

__device__ int floatToOrderedInt( float floatVal )
{
	int intVal = __float_as_int( floatVal );
	return (intVal >= 0 ) ? intVal : intVal ^ 0x7FFFFFFF;
}

__device__ float orderedIntToFloat( int intVal )
{
	return __int_as_float( (intVal >= 0) ? intVal : intVal ^ 0x7FFFFFFF );
}

__global__ void flt_compute_dirixle_case(float *u, float *u_new, float *u_exact, float *f, float eps, int n, int m, float hx, float hy, int *CntIter) 
{
	int i = blockIdx.x * blockDim.x + threadIdx.x; 
 	int j = blockIdx.y * blockDim.y + threadIdx.y;
	float dif;	
	int idx = i*blockDim.y + j;
 	if ((idx > blockDim.x) && (idx % blockDim.x != 0 ) &&
 		((idx + 1) % blockDim.x != 0 ) && (idx < (blockDim.x - 1)*blockDim.x)) {
 		do { 
				exact = 0;				
 				u_new[idx] = 0.25f * (u[idx - 1] + u[idx + 1] + u[idx - n] + u[idx + n] - hx * hy * f[idx]);	 			
				__syncthreads();
 				u[idx] = 0.25f * (u_new[idx - 1] + u_new[idx + 1] + u_new[idx - n] + u_new[idx + n] - hx * hy * f[idx]); 			
				__syncthreads();
				dif = fabs(u[idx] - u_new[idx]);
				atomicMax(&exact, floatToOrderedInt(dif));								   			  			
				*CntIter += 2;
			} while (exact > floatToOrderedInt(eps));
	}
}

__global__ void dbl_compute_dirixle_case(double *u, double *u_new, double *u_exact, double *f, double eps, int n, int m, double hx, double hy, int *CntIter) 
{
	int i = blockIdx.x * blockDim.x + threadIdx.x; 
 	int j = blockIdx.y * blockDim.y + threadIdx.y;
	float dif;	
	int idx = i*blockDim.y + j;
 	if ((idx > blockDim.x) && (idx % blockDim.x != 0 ) &&
 		((idx + 1) % blockDim.x != 0 ) && (idx < (blockDim.x - 1)*blockDim.x)) {
 	do { 
			exact = 0;				
 			u_new[idx] = 0.25f * (u[idx - 1] + u[idx + 1] + u[idx - n] + u[idx + n] - hx * hy * f[idx]);	 			
		    __syncthreads();
 			u[idx] = 0.25f * (u_new[idx - 1] + u_new[idx + 1] + u_new[idx - n] + u_new[idx + n] - hx * hy * f[idx]); 			
		    __syncthreads();
			dif = fabs(u[idx] - u_new[idx]);
			atomicMax(&exact, floatToOrderedInt(dif));								   			  			
			*CntIter += 2;
		} while (exact > floatToOrderedInt(eps));
	}
}

__global__ void flt_test_compute_per_block_u(float *u, float *u_new, float *f, float eps, int n, int m, float hx, float hy)
{

		int tx    = threadIdx.x;       
    int ty    = threadIdx.y;
	int bx    = blockIdx.x;        
    int by    = blockIdx.y;
	int begin = n*(BLOCK_SIZE-1)*by + (BLOCK_SIZE-1)*bx;
	__shared__ float shU_old [BLOCK_SIZE+1][BLOCK_SIZE+1];
	__shared__ float shU_new [BLOCK_SIZE+1][BLOCK_SIZE+1];
	__shared__ float shF     [BLOCK_SIZE+1][BLOCK_SIZE+1];

 	shU_old[ty][tx] = u    [begin + n*ty + tx]; 
 	shU_new[ty][tx] = u_new[begin + n*ty + tx];
	shF[ty][tx]     = f    [begin + n*ty + tx];
    __syncthreads();
	if ((ty > 0) && (ty < BLOCK_SIZE) && (tx > 0) && (tx < BLOCK_SIZE))	
		shU_old[ty][tx] = 0.25f * (shU_new[ty][tx-1] + shU_new[ty][tx+1] + shU_new[ty-1][tx] + shU_new[ty+1][tx] - hx * hy * shF[ty][tx]);					
	
	if ((ty > 0) && (ty < BLOCK_SIZE) && (tx > 0) && (tx < BLOCK_SIZE))
		u[begin + n*ty + tx] = shU_old[ty][tx];			
}

__global__ void flt_test_compute_per_block_unew(float *u, float *u_new, float *f, float eps, int n, int m, float hx, float hy)
{
	 int tx    = threadIdx.x;       
    int ty    = threadIdx.y;
	int bx    = blockIdx.x;        
    int by    = blockIdx.y;

	int begin = n*(BLOCK_SIZE-1)*by + (BLOCK_SIZE-1)*bx;
 
	__shared__ float shU_old [BLOCK_SIZE+1][BLOCK_SIZE+1];
	__shared__ float shU_new [BLOCK_SIZE+1][BLOCK_SIZE+1];
	__shared__ float shF     [BLOCK_SIZE+1][BLOCK_SIZE+1];

	shU_old[ty][tx] = u    [begin + n*ty + tx]; 
	shU_new[ty][tx] = u_new[begin + n*ty + tx];
	shF[ty][tx]     = f    [begin + n*ty + tx];
 	__syncthreads();

	if ((ty > 0) && (ty < BLOCK_SIZE) && (tx > 0) && (tx < BLOCK_SIZE))							
 		shU_new[ty][tx] = 0.25f * (shU_old[ty][tx-1] + shU_old[ty][tx+1] + shU_old[ty-1][tx] + shU_old[ty+1][tx] - hx * hy * shF[ty][tx]);

	if ((ty > 0) && (ty < BLOCK_SIZE) && (tx > 0) && (tx < BLOCK_SIZE))
		u_new[begin + n*ty + tx] = shU_new[ty][tx];	
} 	

__global__	void vectors_dif(float *u1, float *u2, float *sum, int length)
{
	int i = blockIdx.x * blockDim.x + threadIdx.x; 
 	int j = blockIdx.y * blockDim.y + threadIdx.y;
	int tid = i*blockDim.y + j;
	sum[tid] = fabs(u1[tid] - u2[tid]); 	
}


void flt_jacoby_on_device_call(float *u, float *u_new, float *u_exact, float *f, fltParamOfMatrix &POM) 
{   
    float *devU        = NULL;
    float *devU_new    = NULL;
    float *devU_exact  = NULL;
    float *devF        = NULL; 
    int    n           = POM.N;
	int    m           = POM.M;
    float  hx          = POM.hx;
    float  hy          = POM.hy;
    float  eps         = POM.eps;
    int    CountOfIter = 0;    //     1,     atomicInc
    size_t numBytes    = n*m*sizeof(float);
    int  *devCntIter; 
    float  gpuTime     = 0.f;
    cudaEvent_t        start, stop;   

   //     
   dim3 threads(n, m);
   dim3 grid   (1, 1);       

   //        u, u_new   
   cudaMalloc((void**)&devU,        numBytes); //   float   double      
   cudaMalloc((void**)&devU_new,    numBytes);
   cudaMalloc((void**)&devU_exact,  numBytes); 
   cudaMalloc((void**)&devF,        numBytes);
   cudaMalloc((void**)&devCntIter,  sizeof(int));
 
    cudaEventCreate(&start);
    cudaEventCreate(&stop ); 
    cudaEventRecord (start, 0);
    
    //     host   device //   !!!   
    cudaMemcpy(devU,       u,       numBytes, cudaMemcpyHostToDevice); 
    cudaMemcpy(devU_new,   u_new,   numBytes, cudaMemcpyHostToDevice); 
    cudaMemcpy(devU_exact, u_exact, numBytes, cudaMemcpyHostToDevice); 
    cudaMemcpy(devF,       f,       numBytes, cudaMemcpyHostToDevice); 	   

    //   Device 
    flt_compute_dirixle_case<<<grid, threads>>>(devU, devU_new, devU_exact, devF, eps, n, m, hx, hy, devCntIter);   
   
   printf("CUDA error kernel: %s\n ", cudaGetErrorString(cudaGetLastError()));   
	   
   cudaEventRecord      (stop, 0);
   cudaEventSynchronize (stop);

   //   Device 
   cudaMemcpy(u,         devU,    numBytes,    cudaMemcpyDeviceToHost);
   cudaMemcpy(&CountOfIter, devCntIter, sizeof(int), cudaMemcpyDeviceToHost);    
   cudaEventElapsedTime (&gpuTime, start, stop );      
   printf("\n Count of iteration = %d\n\n", CountOfIter);
   printf("\n Time elapsed = %f\n\n", gpuTime);
     
   //    Device
   cudaFree(devF       ); 
   cudaFree(devU       );
   cudaFree(devU_new   );
   cudaFree(devU_exact );  
   cudaFree (devCntIter);
   cudaEventDestroy (start);
   cudaEventDestroy (stop );
   //  
   POM.Result = u; 
   return;   
}

__device__ void warp_reduce_max( float smem[64])
{

	smem[threadIdx.x] = smem[threadIdx.x+32] > smem[threadIdx.x] ? 
						smem[threadIdx.x+32] : smem[threadIdx.x]; __syncthreads();

	smem[threadIdx.x] = smem[threadIdx.x+16] > smem[threadIdx.x] ? 
						smem[threadIdx.x+16] : smem[threadIdx.x]; __syncthreads();

	smem[threadIdx.x] = smem[threadIdx.x+8] > smem[threadIdx.x] ? 
						smem[threadIdx.x+8] : smem[threadIdx.x]; __syncthreads();

	smem[threadIdx.x] = smem[threadIdx.x+4] > smem[threadIdx.x] ? 
						smem[threadIdx.x+4] : smem[threadIdx.x]; __syncthreads();

	smem[threadIdx.x] = smem[threadIdx.x+2] > smem[threadIdx.x] ? 
						smem[threadIdx.x+2] : smem[threadIdx.x]; __syncthreads();

	smem[threadIdx.x] = smem[threadIdx.x+1] > smem[threadIdx.x] ? 
						smem[threadIdx.x+1] : smem[threadIdx.x]; __syncthreads();

}
template<int threads>
__global__ void find_min_max_dynamic(float* in, float* out, int n, int start_adr, int num_blocks)
{	
	__shared__ float smem_max[64];

	int tid = threadIdx.x + start_adr;

	float max = -inf;
	float val;

	// tail part
	int mult = 0;
	for(int i = 1; mult + tid < n; i++)
	{
		val = in[tid + mult];	
		
		max = val > max ? val : max;

		mult = int_mult(i,threads);
	}

	// MAX part
	for(int i = 1; mult+threadIdx.x < num_blocks*2; i++)
	{
		val = out[threadIdx.x + mult];

		max = val > max ? val : max;
		
		mult = int_mult(i,threads);
	}

	if(threads == 32)		
		smem_max[threadIdx.x+32] = 0.0f;	
	smem_max[threadIdx.x] = max;

	__syncthreads();

	if(threadIdx.x < 32)		
		warp_reduce_max(smem_max);
	
	if(threadIdx.x == 0)		
		out[blockIdx.x + gridDim.x] = smem_max[threadIdx.x]; 
	
}

template<int els_per_block, int threads>
__global__ void find_min_max(float* in, float* out, int* indices)
{
	__shared__ float smem_max[64];

	int tid = threadIdx.x + blockIdx.x*els_per_block;

	float max = -inf;	
	float val;	
	int max_index = -1;

	const int iters = els_per_block/threads;

		for(int i = 0; i < iters; i++)
		{

			val = in[tid + i*threads];			
			
			if(val > max)
			{
				max = val;
				max_index = tid+int_mult(i,threads); 
			}
			
		}
	
	
	if(threads == 32)	
		smem_max[threadIdx.x+32] = 0.0f;	
	
	smem_max[threadIdx.x] = max;

	__syncthreads();

	if(threadIdx.x < 32)
		warp_reduce_max(smem_max);	
	
	__syncthreads();

	if(threadIdx.x == 0)		
		out[blockIdx.x + gridDim.x] = smem_max[threadIdx.x]; 
	

	// fix indices	
	if(smem_max[0] == max)	
		indices[blockIdx.x + gridDim.x] = max_index; // MAX
}



void findBlockSize(int* whichSize, int* num_el)
{

	const float pretty_big_number = 24.0f*1024.0f*1024.0f;

	float ratio = float((*num_el))/pretty_big_number;


	if(ratio > 0.8f)
		(*whichSize) =  5;
	else if(ratio > 0.6f)
		(*whichSize) =  4;
	else if(ratio > 0.4f)
		(*whichSize) =  3;
	else if(ratio > 0.2f)
		(*whichSize) =  2;
	else
		(*whichSize) =  1;


}


void compute_reduction(float* d_in, float* d_out, int* d_indices, int num_els)
{

	int whichSize = -1; 
		
	findBlockSize(&whichSize,&num_els);

	int block_size = powf(2,whichSize-1)*blockSize1;
	int num_blocks = num_els/block_size;
	int tail = num_els - num_blocks*block_size;
	int start_adr = num_els - tail;

	
	if(whichSize == 1)
		find_min_max<blockSize1,threads><<< num_blocks, threads>>>(d_in, d_out, d_indices); 
	else if(whichSize == 2)
		find_min_max<blockSize1*2,threads><<< num_blocks, threads>>>(d_in, d_out, d_indices); 
	else if(whichSize == 3)
		find_min_max<blockSize1*4,threads><<< num_blocks, threads>>>(d_in, d_out, d_indices); 
	else if(whichSize == 4)
		find_min_max<blockSize1*8,threads><<< num_blocks, threads>>>(d_in, d_out, d_indices); 
	else
		find_min_max<blockSize1*16,threads><<< num_blocks, threads>>>(d_in, d_out, d_indices); 

	find_min_max_dynamic<threads><<< 1, threads>>>(d_in, d_out, num_els, start_adr, num_blocks);
	
}


void dbl_jacoby_on_device_call(double *u, double *u_new, double *u_exact, double *f, dblParamOfMatrix &POM) 
{   
    double *devU        = NULL;
    double *devU_new    = NULL;
    double *devU_exact  = NULL;
    double *devF        = NULL; 
    int    n            = POM.N;
    int    m            = POM.M;
    double  hx          = POM.hx;
    double  hy          = POM.hy;
    double  eps         = POM.eps;
    int    CountOfIter  = 0;    //     1,     atomicInc
    size_t numBytes     = n*m*sizeof(double);   
	float  gpuTime      = 0.f;
	int  *devCntIter = NULL; 
    cudaEvent_t        start, stop;   

   //     
   dim3 threads(n, m);
   dim3 grid   (1, 1);       

   //        u, u_new   
   cudaMalloc((void**)&devU,        numBytes); //   float   double      
   cudaMalloc((void**)&devU_new,    numBytes);
   cudaMalloc((void**)&devU_exact,  numBytes); 
   cudaMalloc((void**)&devF,        numBytes);
   cudaMalloc((void**)&devCntIter,  sizeof(int));
 
   cudaEventCreate(&start);
   cudaEventCreate(&stop ); 
   cudaEventRecord (start, 0);
    
   //     host   device //   !!!   
   cudaMemcpy(devU,       u,       numBytes, cudaMemcpyHostToDevice); 
   cudaMemcpy(devU_new,   u_new,   numBytes, cudaMemcpyHostToDevice); 
   cudaMemcpy(devU_exact, u_exact, numBytes, cudaMemcpyHostToDevice); 
   cudaMemcpy(devF,       f,       numBytes, cudaMemcpyHostToDevice); 	   

    //   Device 
   dbl_compute_dirixle_case<<<grid, threads>>>(devU, devU_new, devU_exact, devF, eps, n, m, hx, hy, devCntIter);   
   
   printf("CUDA error kernel: %s\n ", cudaGetErrorString(cudaGetLastError()));   
	   
   cudaEventRecord      (stop, 0);
   cudaEventSynchronize (stop   );

   //   Device 
   cudaMemcpy(u,            devU,       numBytes,    cudaMemcpyDeviceToHost);
   cudaMemcpy(&CountOfIter, devCntIter, sizeof(int), cudaMemcpyDeviceToHost);    
   cudaEventElapsedTime (&gpuTime,          start, stop);      
   printf("\n Count of iteration = %d\n\n", CountOfIter);
   printf("\n Time elapsed = %f\n\n",       gpuTime    );
     
   //    Device
   cudaFree(devF       ); 
   cudaFree(devU       );
   cudaFree(devU_new   );
   cudaFree(devU_exact ); 
   cudaFree (devCntIter);
   cudaEventDestroy (start);
   cudaEventDestroy (stop );
   //  
   POM.Result = u; 
   return;
}

void test_compute_per_blocks(float *u, float *u_new, float *u_exact, float *f, fltParamOfMatrix &POM)
{
    float *devU        = NULL;
    float *devU_new    = NULL;    
    float *devF        = NULL; 
    int    n           = POM.N;
	int    m           = POM.M;
    float  hx          = POM.hx;
    float  hy          = POM.hy;
    float  eps         = POM.eps;
	// for get max value 
	//--------------------
	
	int size = n*m*sizeof(float);
	int index_size = 2048*sizeof(int); // max 2048 blocks! Bug warning!!
	float *sum = (float*) malloc(size);	
	int* d_indices;	
	float* d_out;	
	cudaMalloc((void**)&d_out, size);
	cudaMalloc((void**)&d_indices, index_size);
	cudaMalloc((void**)&sum,   size);    
	//--------------------
	
	float *devExact   = NULL;
	float *exactArray = NULL;
		
	int grid_width = 1;
	int grid_heigth = 1;
	int CntOfIter = 0; 
   	float  gpuTime     = 0.f;

   	cudaEvent_t        start, stop;  	

	//     
   dim3 threads(BLOCK_SIZE + 1, BLOCK_SIZE + 1);

   if ((n - 2*BLOCK_SIZE) == 0){
	   grid_width = ceil((float)n/BLOCK_SIZE);
	   grid_heigth = ceil((float)n/BLOCK_SIZE);
   }
   else 
	   if ( (n - 2*BLOCK_SIZE) % BLOCK_SIZE*(BLOCK_SIZE-1) == 0) {
		grid_width = (int) (ceil((float)n/BLOCK_SIZE) + (n - 2*BLOCK_SIZE)/(BLOCK_SIZE*(BLOCK_SIZE-1)));
		grid_heigth = (int) (ceil((float)n/BLOCK_SIZE) + (n - 2*BLOCK_SIZE)/(BLOCK_SIZE*(BLOCK_SIZE-1)));
	   }
   dim3 grid   (grid_width, grid_heigth);  

   //        u, u_new   
   cudaMalloc((void**)&devU,        size); 
   cudaMalloc((void**)&devU_new,    size);   
   cudaMalloc((void**)&devF,        size);       
   cudaMalloc((void**)&devExact, grid_width*grid_heigth*sizeof(float));
	
   cudaEventCreate(&start);
   cudaEventCreate(&stop );    
   cudaMemcpy(devU,       u,       size, cudaMemcpyHostToDevice); 		
   cudaMemcpy(devU_new,   u_new,   size, cudaMemcpyHostToDevice);	  
   cudaMemcpy(devF,       f,       size, cudaMemcpyHostToDevice); 
   
   cudaEventRecord (start, 0);
   do {
	for (int k = 0; k < CntRunKernel; ++k) {
		flt_test_compute_per_block_unew<<<grid, threads>>>(devU, devU_new, devF, eps, n, m, hx, hy);			
		flt_test_compute_per_block_unew<<<grid, threads>>>(devU_new, devU, devF, eps, n, m, hx, hy);  						
	}
	vectors_dif<<<grid, threads>>>(devU, devU_new, sum, n*m);	
	compute_reduction(sum, d_out, d_indices, n*m);	
	cudaMemcpy(sum, d_out, n*m, cudaMemcpyDeviceToHost);
	
	CntOfIter += 2*CntRunKernel;
   } while (sum[0] > eps);

   cudaEventRecord      (stop, 0);
   cudaEventSynchronize (stop);
   cudaEventElapsedTime (&gpuTime, start, stop );  
	//  
   cudaMemcpy(u,             devU,        size,    cudaMemcpyDeviceToHost);    

   FILE *fp = fopen("Result.txt", "w");
   for (int i = 0; i < n; ++i) {
	   for (int j = 0; j < m; ++j)	
		   fprintf(fp, "%0*.*f\t",1,6, u[n*i + j]);	
	   fprintf(fp, "\n");
   }
   fclose(fp);

   fp = fopen("Iteration_Time.log", "a");
   fprintf(fp, "N\t M\t hx\t\t\t hy\t\t\t\t eps\t\t\t gpuTime\t CntOfIter\t\n", n);
   fprintf(fp, "%d\t ", n);
   fprintf(fp, "%d\t ", m);
   fprintf(fp, "%f\t ", hx);
   fprintf(fp, "%f\t\t ", hy);
   fprintf(fp, "%f\t\t ", eps);
   fprintf(fp, "%f\t\t ", gpuTime);
   fprintf(fp, "%d\n\n", CntOfIter);
   fclose(fp);

   cudaFree(devF       ); 
   cudaFree(devU       );
   cudaFree(devU_new   );
   cudaFree(sum);
   cudaFree(devExact);
   cudaFree(d_out);
   cudaFree(d_indices);
   free(exactArray);
   cudaEventDestroy (start);
   cudaEventDestroy (stop );
   return;
}

